#ifndef _DISTANCE_HPP_
#define _DISTANCE_HPP_

#if defined SSE && defined __SSE__
	#include <immintrin.h>
#endif

#include <opencv2/imgproc/imgproc.hpp>

namespace histodet {

/**
 * Abstract class for distance between two histograms.
 * This class is actually not used because Distance classes must be used as templates
 * for performance reasons, but it describes the interface the Distance classes must provide.
 *
 * Some distances are inverted, meaning that higher value means closer (eg correlation).
 *
 * Distances are not normalized, so they offer to transform the threshold instead
 * of transforming all values for optimal performance, and to compare a value to the threshold
 * (depending on if it is inverted or not).
 */
struct DistanceAbstract
{
	/**
	 * Compute distance between two histograms, optionally using the histogram standard deviation
	 * @param hist1 first histogram
	 * @param hist2 second histogram
	 * @param hist_sig standard deviations of histogram
	 * @return distance between the two histograms
	 */
	static inline float dist(cv::Mat const& hist1, cv::Mat const& hist2, cv::Mat const& hist_sig) { return 0.; }
	/**
	 * Transform the normalized threshold between 0 and 1 into a threshold that is comparable
	 * with the distance output.
	 * @param thres the normalized threshold between 0 and 1
	 * @param area the number of pixels involved in the histogram
	 * @param hist_size the number of bins of the histogram
	 * @param[out] dist_inverted returns whether this distance is inverted or not,
	 * in order to know how to compare a distance with the threshold
	 * @return the equivalent threshold comparable with the distance value
	 */
	static inline float transform_threshold(float thres, float area, float hist_size, bool &dist_inverted) { return 0.; }
	/**
	 * Normalize the distance value between 0 and 1.
	 * Not efficient, only used for display and analysis.
	 * For normal use, transform the threshold instead.
	 * @param dist the distance raw result
	 * @param area the number of pixels involved in the histogram
	 * @param hist_size the number of bins of the histogram
	 * @return the normalized distance value between 0 and 1
	 */
	static inline float normalize(float dist, float area, float hist_size) { return 0.;}
	/**
	 * Compare a distance value with a threshold, depending on whether it is inverted
	 * or not. No overhead thanks to templates and inlining.
	 * @param value the distance value
	 * @param threshold the transformed threshold
	 * @returns whether the distance value is better than the threshold (positive result)
	 */
	static inline bool compare(float value, float thres) { return false; }
	/**
	 * @returns the worst possible distance (depending on whether it is inverted or not)
	 */
	static inline float infinite() { return 0.; }
};


template<typename T>
inline T sqr(T x) { return x*x; }

/**
 * Correlation distance
 */
struct DistanceCorrel
{
	static inline float dist(cv::Mat const& hist1, cv::Mat const& hist2, cv::Mat const& hist_sig)
		{ return compareHist(hist1, hist2, CV_COMP_CORREL); }
	static inline float transform_threshold(float thres, float area, float hist_size, bool &dist_inverted)
		{ dist_inverted = true; return 1-thres; }
	static inline float normalize(float dist, float area, float hist_size)
		{ return 1-dist; }
	static inline bool compare(float value, float thres)
		{ return value >= thres; } // inverted
	static inline float infinite()
		{ return 0.0; } // inverted
};


/**
 * Chi square distance
 */
struct DistanceChiSqr
{
	static inline float dist(cv::Mat const& hist1, cv::Mat const& hist2, cv::Mat const& hist_sig)
		{ return compareHist(hist1, hist2, CV_COMP_CHISQR); }
	static inline float transform_threshold(float thres, float area, float hist_size, bool &dist_inverted)
		{ dist_inverted = false; return thres * (2.0 * area); }
	static inline float normalize(float dist, float area, float hist_size)
		{ return dist / (2.0 * area); }
	static inline bool compare(float value, float thres)
		{ return value <= thres; } // not inverted
	static inline float infinite()
		{ return 1e10; } // not inverted
};


/**
 * Intersection distance
 */
struct DistanceInter
{
	static inline float dist(cv::Mat const& hist1, cv::Mat const& hist2, cv::Mat const& hist_sig)
		{ return compareHist(hist1, hist2, CV_COMP_INTERSECT); }
	static inline float transform_threshold(float thres, float area, float hist_size, bool &dist_inverted)
		{ dist_inverted = true; return (1-thres) * area; }
	static inline float normalize(float dist, float area, float hist_size)
		{ return 1 - (dist/area); }
	static inline bool compare(float value, float thres)
		{ return value >= thres; } // inverted
	static inline float infinite()
		{ return 0.0; } // inverted
};


/**
 * L1 distance
 */
struct DistanceL1
{
	static inline float dist(cv::Mat const& hist1, cv::Mat const& hist2, cv::Mat const& hist_sig)
		{ return cv::norm(hist1,hist2, cv::NORM_L1); }
	static inline float transform_threshold(float thres, float area, float hist_size, bool &dist_inverted)
		{ dist_inverted = false; return thres * (2.0 * area); }
	static inline float normalize(float dist, float area, float hist_size)
		{ return dist / (2.0 * area); }
	static inline bool compare(float value, float thres)
		{ return value <= thres; } // not inverted
	static inline float infinite()
		{ return 1e10; } // not inverted
};


/**
 * Normalized L1 distance
 */
struct DistanceL1n
{
	static inline float dist(cv::Mat const& hist1, cv::Mat const& hist2, cv::Mat const& hist_sig)
	{
	#if defined SSE && defined __SSE__
			// This loop is NOT automatically vectorized by GCC because it is too complicated
			// (try gcc option -ftree-vectorizer-verbose)
			union { __m128 v; float f[4]; } sum = {_mm_setzero_ps()};
			for(int i = 0; i < hist1.rows; i+=4)
				sum.v = _mm_add_ps(sum.v, _mm_and_ps(_mm_div_ps(_mm_sub_ps(*((__m128*)((float*)hist1.data+i)), *((__m128*)((float*)hist2.data+i))), *((__m128*)((float*)hist_sig.data+i))), sign_mask.v));
			return (sum.f[0]+sum.f[1]+sum.f[2]+sum.f[3]);
	#elif defined OPENCV
			return cv::norm((hist1-hist2)/hist_sig, cv::NORM_L1);
	#else
			float res = 0.;
			for(int i = 0; i < hist1.rows; ++i)
				res += fabsf((((float*)hist1.data)[i] - ((float*)hist2.data)[i]) / ((float*)hist_sig.data)[i]);
			return res;
	#endif
		}
	static inline float transform_threshold(float thres, float area, float hist_size, bool &dist_inverted)
		{ dist_inverted = false; return thres * hist_size * 2.0; } // may overflow
	static inline float normalize(float dist, float area, float hist_size)
		{ float coef = hist_size * 2.0; if (dist > coef) dist = coef; return dist / coef; }
	static inline bool compare(float value, float thres)
		{ return value <= thres; } // not inverted
	static inline float infinite()
		{ return 1e10; } // not inverted
};

/**
 * Normalized positive distance (only adds positive differences, 0 otherwise)
 */
struct DistancePOSn
{
	static inline float dist(cv::Mat const& hist1, cv::Mat const& hist2, cv::Mat const& hist_sig)
	{
	#if defined SSE && defined __SSE__
			// This loop is NOT automatically vectorized by GCC because it is too complicated
			// (try gcc option -ftree-vectorizer-verbose)
			union { __m128 v; float f[4]; int i[4]; } diff;
			float sum = 0.0;
			for(int i = 0; i < hist1.rows; i+=4)
			{
				diff.v = _mm_div_ps(_mm_sub_ps(*((__m128*)((float*)hist2.data+i)), *((__m128*)((float*)hist1.data+i))), *((__m128*)((float*)hist_sig.data+i)));
				if (diff.i[0] > 0) sum += diff.f[0]; // comparing diff.i makes the whole detection take 8.5ms and diff.f 10ms !
				if (diff.i[1] > 0) sum += diff.f[1];
				if (diff.i[2] > 0) sum += diff.f[2];
				if (diff.i[3] > 0) sum += diff.f[3];
			}
			return sum;
	#elif defined OPENCV
			return 0.0; // not implemented in opencv
	#else
			float sum = 0., diff;
			for(int i = 0; i < hist1.rows; ++i)
			{
				diff = ((float*)hist2.data)[i] - ((float*)hist1.data)[i];
				if (*(int*)(&diff) > 0) sum += diff / ((float*)hist_sig.data)[i];
			}
			return sum;
	#endif
		}
	static inline float transform_threshold(float thres, float area, float hist_size, bool &dist_inverted)
		{ dist_inverted = false; return thres * hist_size * 2.0; } // may overflow
	static inline float normalize(float dist, float area, float hist_size)
		{ float coef = hist_size * 2.0; if (dist > coef) dist = coef; return dist / coef; }
	static inline bool compare(float value, float thres)
		{ return value <= thres; } // not inverted
	static inline float infinite()
		{ return 1e10; } // not inverted
};


/**
 * L2 distance
 */
struct DistanceL2
{
	static inline float dist(cv::Mat const& hist1, cv::Mat const& hist2, cv::Mat const& hist_sig)
		{ return cv::norm(hist1,hist2, cv::NORM_L2); }
	static inline float transform_threshold(float thres, float area, float hist_size, bool &dist_inverted)
		{ dist_inverted = false; return thres * (2.0 * area); }
	static inline float normalize(float dist, float area, float hist_size)
		{ return dist / (2.0 * area); }
	static inline bool compare(float value, float thres)
		{ return value <= thres; } // not inverted
	static inline float infinite()
		{ return 1e10; } // not inverted
};


/**
 * Normalized L2 distance
 */
struct DistanceL2n
{
	static inline float dist(cv::Mat const& hist1, cv::Mat const& hist2, cv::Mat const& hist_sig)
	{
	#if defined SSE && defined __SSE__
			// This loop is NOT automatically vectorized by GCC because it is too complicated
			// (try gcc option -ftree-vectorizer-verbose)
			union { __m128 v; float f[4]; } sum = {_mm_setzero_ps()};
			__m128 e;
			for(int i = 0; i < hist1.rows; i+=4)
			{
				e = _mm_div_ps(_mm_sub_ps(*((__m128*)((float*)hist1.data+i)), *((__m128*)((float*)hist2.data+i))), *((__m128*)((float*)hist_sig.data+i)));
				sum.v = _mm_add_ps(sum.v, _mm_mul_ps(e,e));
			}
// 			sum.v = _mm_hadd_ps(sum.v, sum.v); // uses SSE3 and not really faster
// 			sum.v = _mm_hadd_ps(sum.v, sum.v);
// 			return sum.f[0];
			return (sum.f[0]+sum.f[1]+sum.f[2]+sum.f[3]);
	#elif defined OPENCV
			return cv::norm((hist1-hist2)/hist_sig, cv::NORM_L2);
	#else
			float res = 0.;
			for(int i = 0; i < hist1.rows; ++i)
			{
				float e = (((float*)hist1.data)[i] - ((float*)hist2.data)[i]) / ((float*)hist_sig.data)[i];
				res += e*e;
			}
			return res;
	#endif
	}
	static inline float transform_threshold(float thres, float area, float hist_size, bool &dist_inverted)
	{
	#if defined OPENCV
		dist_inverted = false; return thres * 10.0;
	#else
		dist_inverted = false; return sqr(thres * 10.0); // 1 = 10 sigma, may overflow
	#endif
	}
	static inline float normalize(float dist, float area, float hist_size)
	{
	#if defined OPENCV
		return dist / 10.0;
	#else
		return sqrt(dist) / 10.0; // 1 = 10 sigma, may overflow
	#endif
	}
	static inline bool compare(float value, float thres)
		{ return value <= thres; } // not inverted
	static inline float infinite()
		{ return 1e10; } // not inverted
};


/**
 * Infinite distance
 */
struct DistanceInf
{
	static inline float dist(cv::Mat const& hist1, cv::Mat const& hist2, cv::Mat const& hist_sig)
		{ return cv::norm(hist1,hist2, cv::NORM_INF); }
	static inline float transform_threshold(float thres, float area, float hist_size, bool &dist_inverted)
		{ dist_inverted = false; return thres * (2.0 * area); }
	static inline float normalize(float dist, float area, float hist_size)
		{ return dist / (2.0 * area); }
	static inline bool compare(float value, float thres)
		{ return value <= thres; } // not inverted
	static inline float infinite()
		{ return 1e10; } // not inverted
};


} // namespace histodet


#endif
