#ifndef _HISTOGRAM_DISTRIBUTIVE_HPP_
#define _HISTOGRAM_DISTRIBUTIVE_HPP_

#if defined SSE && defined __SSE__
	#include <immintrin.h>
#endif
#include "histodet/histogramAbstract.hpp"


namespace histodet {

	
/**
 * Computes histogram of a sliding window in an image with the distributive
 * histogram algorithm.
 * See documentation of HistogramAbstract for interface description
 */
template<typename imgtype>
class HistogramDistributive: public HistogramAbstract
{
 private:
	int ncols;
	Hist *col_hists;
	imgtype *img_prev_row, *img_next_row;
	Hist *hist_next_col;

 private:
	inline void updateCol()
	{
		hist_next_col->data()[*(img_prev_row)] -= 1.;
		hist_next_col->data()[*(img_next_row)] += 1.;
		++hist_next_col, ++img_prev_row, ++img_next_row;
	}

 public:
	HistogramDistributive(int hist_size, BucketerAbstract *bucketer);
	~HistogramDistributive();
	void loadImage(cv::Mat const &image, int window_size);
	inline cv::Mat& get(cv::Rect const& rect);
	inline cv::Mat& getNext(int &pos_x, int &pos_y);
};



///*********************************************************************************************************************************************
/// Implementation
///*********************************************************************************************************************************************

template<typename imgtype>
HistogramDistributive<imgtype>::HistogramDistributive(int hist_size, BucketerAbstract *bucketer):
	HistogramAbstract(hist_size, bucketer), ncols(0), col_hists(NULL)
{}

template<typename imgtype>
HistogramDistributive<imgtype>::~HistogramDistributive()
{
	delete[] col_hists;
}

template<typename imgtype>
cv::Mat& HistogramDistributive<imgtype>::get(cv::Rect const& rect)
{
	hist.clear();
	for(int y = rect.y; y < rect.y+rect.height; ++y)
	for(int x = rect.x; x < rect.x+rect.width; ++x)
		hist.data()[img.ptr<imgtype>(y)[x]] += 1.;

	return hist;
}



template<typename imgtype>
cv::Mat& HistogramDistributive<imgtype>::getNext(int &pos_x, int &pos_y)
{
#ifndef NDEBUG
	if (!img.data) { std::cout << "Please load image before trying to compute histograms" << std::endl; return hist; }
#endif

	if (first_col)
	{
		if (finished) { pos_x = pos_y = -1; return hist; } // optimal to do it here!
		if (!first_line) // update column hists
		{
			img_prev_row = (imgtype*)(img.data + img.step[0]*(cur_y-win_hsize-1));
			img_next_row = (imgtype*)(img.data + img.step[0]*(cur_y+win_hsize  ));
			hist_next_col = col_hists;
			for(int x = 0; x < win_size; ++x) updateCol();
		}

		first_col = false;
		hist.clear();
#if defined SSE && defined __SSE__
		// This loop is automatically vectorized by GCC but less efficiently because
		// it performs tests for aliasing and size
		for(int x = 0; x < win_size; ++x)
		{
			float *cur = hist.data(), *ch = col_hists[x].data();
			for(int i = 0; i < hist_size; i+=4)
				*((__m128*)(cur+i)) = _mm_add_ps(*((__m128*)(cur+i)), *((__m128*)(ch+i)));
		}
#elif defined OPENCV
		for(int x = 0; x < win_size; ++x) hist() += col_hists[x]();
#else
		for(int x = 0; x < win_size; ++x)
		{
			float *cur = hist.data(), *ch = col_hists[x].data();
			for(int i = 0; i < hist_size; ++i) cur[i] += ch[i];
		}
#endif
	} else
	{
		if (!first_line) // update column hists
			updateCol();

#if defined SSE && defined __SSE__
		// This loop is automatically vectorized by GCC but less efficiently because
		// it performs tests for aliasing and size
		float *cur = hist.data(), *prev = col_hists[cur_x-win_hsize-1].data(), *next = col_hists[cur_x+win_hsize].data();
		for(int i = 0; i < hist_size; i+=4)
			*((__m128*)(cur+i)) = _mm_add_ps(*((__m128*)(cur+i)), _mm_sub_ps(*((__m128*)(next+i)), *((__m128*)(prev+i))));
#elif defined OPENCV
		hist() = hist() + col_hists[cur_x+win_hsize]() - col_hists[cur_x-win_hsize-1]();
#else
		float *cur = hist.data(), *prev = col_hists[cur_x-win_hsize-1].data(), *next = col_hists[cur_x+win_hsize].data();
		for(int i = 0; i < hist_size; ++i) cur[i] += next[i] - prev[i];
#endif
	}

	updatePos(pos_x, pos_y);
	return hist;
}



template<typename imgtype>
void HistogramDistributive<imgtype>::loadImage(cv::Mat const &image, int window_size)
{
	bool winsize_changed = (window_size/2 != win_hsize);
	HistogramAbstract::loadImage(image, window_size);

	if (ncols != img.cols || winsize_changed)
	{
		delete[] col_hists;
		ncols = img.cols;
		col_hists = new Hist[ncols];
		for(int x = 0; x < ncols; ++x) col_hists[x].build(hist_size);
	} else
	{
		for(int x = 0; x < ncols; ++x) col_hists[x].clear();
	}

	// initialize column histograms with first line
	// compute here rather than in getNext to make the latter smaller and more inlinable
	for(int x = 0; x < img.cols; ++x)
		for(int y = 0; y < window_size; ++y)
			col_hists[x].data()[*((imgtype*)(img.data+img.step[0]*y)+x)] += 1.;
}


} // namespace histodet

#endif

